// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#include <iostream>

#include <anna/core/core.hpp>

#include <anna/comm/Application.hpp>
#include <anna/comm/Communicator.hpp>

#include <anna/app/functions.hpp>

#include <anna/timex/timex.hpp>

#include <anna/ldap/Engine.hpp>
#include <anna/ldap/Session.hpp>
#include <anna/ldap/Search.hpp>
#include <anna/ldap/Response.hpp>
#include <anna/ldap/Attribute.hpp>

class MySession : public ldap::Session {
   void eventResponse (const ldap::Response&) throw (RuntimeException);
};

class MyEngine : public ldap::Engine {
public:
   MyEngine () {;}

private:
   Recycler<MySession> a_sessions;

   ldap::Session* allocateSession (const int category) throw () { return a_sessions.create (); }

   void releaseSession (ldap::Session* session) throw () { 
      MySession* aux = static_cast <MySession*> (session);
      a_sessions.release (aux);
   }
};

//-----------------------------------------------------------------------------------------
// Define el comunicador de nuestra aplicacin.
//
// Las peticiones y respuestas van codificadas mediante un RawCodec pero podriamos 
// haber utilizado cualquier otro medio de codificacin ya que la capa de transporte
// es totalmente independiente del contenido del mensaje.
//
// De cara a  la capa de transporte lo nico que importa es el cliente y el servidor
// codifiquen/decodifiquen de la misma forma (RawCodec, EnhancedCodec ....)
//-----------------------------------------------------------------------------------------
class MyCommunicator : public comm::Communicator {
public:   
   MyCommunicator () : comm::Communicator () { ;}

private:
   void eventReceiveMessage (comm::ClientSocket &, const comm::Message&) throw (RuntimeException) {;}
};

class Buddy : public Clock {
public:
   Buddy () :  Clock ("buddy", (Millisecond)2000) {;}
   
   void setLDAPEngine (ldap::Engine& ldapEngine) throw () { a_ldapEngine = &ldapEngine;  }
   
private:
   ldap::Engine* a_ldapEngine;   
   ldap::Search a_ldapSearch;
   
   bool tick () throw (RuntimeException);
};

class Stopper : public Timer {
public:
   Stopper () :  Timer ("stopper", (Millisecond)0) {;}
   void expire (Engine*) throw (RuntimeException);      
};

class LDAPClient : public anna::comm::Application {
public:
   LDAPClient ();
   
private:
   MyCommunicator a_communicator;
   anna::timex::Engine a_timeController;
   MyEngine a_ldapEngine;
   Buddy a_buddy;
   Stopper a_stopper;  

   void run () throw (RuntimeException);    
};

using namespace std;
using namespace anna::comm;

int main (int argc, const char** argv)
{
   CommandLine& commandLine (CommandLine::instantiate ());
   LDAPClient app;
   
   try {
      commandLine.initialize (argv, argc);
      commandLine.verify ();

      Logger::setLevel (Logger::Debug); 
      Logger::initialize ("ldap_tsearch", new TraceWriter ("file.trace", 2048000));
 
      app.start ();
   }
   catch (Exception& ex) {
      cout << ex.asString () << endl;
   }
   
   return 0;
}

LDAPClient::LDAPClient () : 
   Application ("ldap_tsearch", "Client LDAP", "1.0"),
   a_communicator (),
   a_timeController ((Millisecond)120000, (Millisecond)1000)
{
   CommandLine& cl = CommandLine::instantiate ();
   
   cl.add ("url", CommandLine::Argument::Mandatory, "URL del host");
   cl.add ("base", CommandLine::Argument::Mandatory, "Base");
   cl.add ("t", CommandLine::Argument::Mandatory, "Segundos que deber estar en ejecucion este cliente");
   cl.add ("filter", CommandLine::Argument::Optional, "Filtro");
   cl.add ("scope", CommandLine::Argument::Optional, "Ambito");
   cl.add ("user", CommandLine::Argument::Optional, "Usuario");
   cl.add ("password", CommandLine::Argument::Optional, "Password");
   cl.add ("d", CommandLine::Argument::Optional, "Nivel de depuracion (all = activa todo)");
}

//-----------------------------------------------------------------------------------------
// Atiende las peticiones. 
// Cuando hay un nuevo mensaje invocar a Communicator::eventReceiveMessage
// 
// (1) Para evitar la ejecucion de la aplicacion en caso de indicar un valor no valido.
//-----------------------------------------------------------------------------------------
void LDAPClient::run ()
   throw (RuntimeException)
{
   LOGMETHOD (TraceMethod tm ("LDAPClient", "run", ANNA_FILE_LOCATION));

   CommandLine& cl (CommandLine::instantiate ());
   
   if (cl.exists ("scope"))
      ldap::Scope::asEnum (cl.getValue ("scope"));  // (1)
   
   if (cl.exists ("d")) {
      if (anna_strcmp (cl.getValue ("d"), "all") == 0)
         ldap::Engine::setDebugLevel (ldap::Engine::DebugLevel::All);
      else                
         ldap::Engine::setDebugLevel (cl.getIntegerValue ("d"));
   }   

   a_buddy.setLDAPEngine (a_ldapEngine);

   a_stopper.setTimeout ((Millisecond)(CommandLine::instantiate ().getIntegerValue ("t") * 1000));

   a_timeController.activate (a_buddy);
   a_timeController.activate (a_stopper);   

   a_communicator.accept ();
}

void MySession::eventResponse (const ldap::Response& response) 
   throw (RuntimeException)
{
   cout << "LDAP Response: " << response.asString () << endl;
   cout << "   Name: " << response.getName () << endl;   

   ldap::Response::const_attribute_iterator ii, maxii;   
   ldap::Attribute::const_value_iterator vv, maxvv;
   const ldap::Attribute* attr;

   for (ii = response.attribute_begin (), maxii = response.attribute_end (); ii != maxii; ii ++) {
      attr = ldap::Response::attribute (ii);

      cout << "   Attribute: " << attr->getName () << " -> ";
      for (vv = attr->value_begin (), maxvv = attr->value_end (); vv != maxvv; vv ++)
         cout << ldap::Attribute::value (vv) << " ";
      cout << endl;
   }

   ldap::Response::const_referral_iterator jj, maxjj;
   for (jj = response.referral_begin (), maxjj = response.referral_end (); jj != maxjj; jj ++)
      cout << "   Referral: " << ldap::Response::referral (jj) << endl;
      
   cout << endl;
}

bool Buddy::tick () 
   throw (RuntimeException) 
{
   CommandLine& cl = CommandLine::instantiate ();

   ldap::Session* session;

   if (cl.exists ("user") && cl.exists  ("password"))
      session = a_ldapEngine->createSession (cl.getValue ("url"), cl.getValue ("user"), cl.getValue ("password"));
   else
      session = a_ldapEngine->createSession (cl.getValue ("url"));
      
   if (session->isBound ()) {
      a_ldapSearch.clear ();
      a_ldapSearch.setBase (cl.getValue ("base"));      
      if (cl.exists ("filter"))
         a_ldapSearch.setFilter (cl.getValue ("filter"));      
      if (cl.exists ("scope"))
         a_ldapSearch.setScope (cl.getValue ("scope"));
      session->setTimeout (ldap::ClassCode::Search, (Millisecond)5000);
      session->send (a_ldapSearch);
   }
   else
      session->bind ();

   return true;
}

void Stopper::expire (Engine*) 
   throw (RuntimeException)
{
   LOGMETHOD (TraceMethod tm ("Stopper", "expire", ANNA_FILE_LOCATION));
   
   MyCommunicator* communicator = anna::app::functions::component <MyCommunicator> (ANNA_FILE_LOCATION);
   
   communicator->requestStop ();
}

