// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_core_AutoPointer_hpp
#define anna_core_AutoPointer_hpp

#include <anna/core/util/String.hpp>

namespace anna {

/**
 * Clase de la que deben heredar todas las clases con capacidades de liberacin automtica.
 * Respector a \em auto_ptr tiene la ventaja de que puede actuar sobre objetos creados en
 * la pila.
 *
 * La clase que implemente este interface debe declara como \em friend a esta clase.
 */
class AutoPointer {
public:
  /**
   * peracin a realizar cuando se termine de tratar con esta instancia y se invoque a anna::AutoPointer::release.
   * \see AutoPointer.
   */
  struct WhenFinished  {
    enum _v {
      Ignore /** << La invocacin al mtodo anna::AutoPointer::release no tendr ningn efecto. */,
      Delete  /** << La invocacin al mtodo anna::AutoPointer::release originar la llamada al operador \em delete de la instancia. */
    };
  };

  /**
   * Destructor.
   */
  virtual ~AutoPointer() {;}

  /**
   * Devuelve la operacin que se realizar cuando se termine con esta instancia y se invoque a anna::AutoPointer::release.
   * \return la operacin que se realizar cuando se termine con esta instancia y se invoque a anna::AutoPointer::release.
   */
  WhenFinished::_v getWhenFinished() const throw() { return a_whenFinished; }

  /**
   * Establece la operacin a realizar cuando se invoca al mtodo anna::AutoPointer::release.
   * \param whenFinished Indica que operacin a realizar cuando se invoque a anna::AutoPointer::release.
   */
  void setWhenFinished(const  WhenFinished::_v whenFinished) throw() { a_whenFinished = whenFinished; }

  /**
     Devuelve una cadena con la informacion relevante de este objeto.
     \return Una cadena con la informacion relevante de este objeto.
  */
  virtual String asString() const throw();

  /**
   * Si fuera necesario libera de la instancia recibida.
   * \param instance Instancia a liberar si fuera necesario. Puede ser NULL.
   * \return Devolver \em NULL si la instancia ha sido liberada o la misma instancia recibida en caso de que no haya sido liberada.
   */
  template <typename T> static T* release(T& instance)
  throw() {
    return release(&instance);
  }

  /**
   * Si fuera necesario libera de la instancia recibida.
   * \param instance Instancia a liberar si fuera necesario. Puede ser NULL.
   * \return Devolver \em NULL si la instancia ha sido liberada o la misma instancia recibida en caso de que no haya sido liberada.
   */
  template <typename T> static T* release(T* instance)
  throw() {
    if(instance == NULL)
      return NULL;

    if(instance->getWhenFinished() == WhenFinished::Delete) {
      delete instance;
      instance = NULL;
    }

    return instance;
  }

  /**
   * Devuelve el nombre lgico de esta clase.
   * \return el nombre lgico de esta clase.
   */
  static const char* className() throw() { return "AutoPointer"; }

protected:
  /**
   * Constructor
   * \param whenFinished Indica que operacin a realizar cuando se invoque a anna::AutoPointer::release.
   */
  AutoPointer(const WhenFinished::_v whenFinished) : a_whenFinished(whenFinished) {;}

private:
  WhenFinished::_v a_whenFinished;
};

}

#endif


