// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_xml_Parser_hpp
#define anna_xml_Parser_hpp

#include <anna/core/RuntimeException.hpp>
#include <anna/core/mt/Mutex.hpp>

struct _xmlValidCtxt;
struct _xmlDoc;
struct _xmlNode;
struct _xmlAttr;

namespace anna {

class DataBlock;

namespace xml {

class Node;
class DTD;
class Document;
class Attribute;

/**
   Analizador de documentos XML.

   Analiza la expresion contenida en un documento XML, opcionalmente puede validar la estructura
   mediante una DTD suministrada a tal efecto, y devuelve un arbol de nodos que resulta muy
   facil de usar.

   Ejemplo de documento XML:

   \code

   <xvc HeartBeat="20000">
      <broadcast>
         <INetAddress Address="204.152.65.15" Port="2000"/>
         <INetAddress Address="204.152.65.47" Port="2002"/>
      </broadcast>

      <ethernet Mode="raw" VirtualAddress="1.1.1.100">
         <Input Device="/dev/qfe" PhysicalAccessPoint="2" MAC="8:0:20:9e:ee:21"/>
         <Output Device="/dev/qfe" PhysicalAccessPoint="2" MAC="8:0:20:9e:ee:c9"/>
      </ethernet>
   </xvc>

   \endcode

   Donde los nodos (Ver \ref Node) \em broadcast y \em ethernet son hijos del nodo \em xvc. De la misma forma
   los nodos \em Input y \em Output son hijos de \em ethernet.

   Cada nodo XML puede tener una serie indeterminada de atributos (Ver \ref Attribute) que completan su
   significado. Por ejemplo \em Device, \em PhysicalAccessPoint y \em MAC son atributos de los nodos \em Input y
   \em Output.
*/
class Parser : public Mutex {
public:
  /**
     Constructor.
  */
  Parser();

  /**
     Destructor.
  */
  virtual ~Parser();

  /**
     Devuelve el nodo raiz de la expresion XML analizada. Puede ser NULL.

     \return El nodo raiz de la expresion XML analizada. Puede ser NULL.
     \warning
        \li El resultado de este metodo solo sera valido despues de invocar a alguno
     de los metodo #apply.
        \li El nodo devuelto  no puede ser usado despues del invocar a destructor de este Parser.
  */
  const Node* getRoot() const throw() { return a_root; }

  /**
     Analiza el documento XML recibido como parametro.
     \param document Documento XML que deseamos analizar.
     \return El nodo raiz del arbol XML correspondiente al documento analizado.
     \warning
        \li La instancia del documento debe estar correctamente inicializada.
        \li El nodo devuelto  no puede ser usado despues del invocar a destructor de este Parser.
     \see Document::initialize
  */
  const Node* apply(const Document& document) throw(RuntimeException);

  /**
     Analiza el archivo XML recibido como parametro, y verifica que cumpla las reglas
     establecidas por la DTD.
     \param document Documento XML que deseamos analizar.
     \param dtd DTD que debe cumplir el documento XML.
     \return El nodo raiz del arbol XML correspondiente al resultado del analisis.

     \warning
        \li La instancia de la DTD y el documento deben estar correctamente inicializadas.
        \li El nodo devuelto  no puede ser usado despues del invocar a destructor de este Parser.

     \see DTD::initialize
     \see Document::initialize
  */
  const Node* apply(const Document& document, const DTD& dtd) throw(RuntimeException);

protected:
  /*
   * Nodo riz del rbol representado por el documento XML analizado.
   */
  Node* a_root;

  /**
     Inicializa el contenido del nodo asociado al ultimo Parser::apply.
     \warning Se invoca automaticamente desde el numero de ANNA.xml
  */
  virtual void reset() throw();

  /**
   * Comprueba si se requieren operaciones de traduccin del conjunto de caracteres.
   */
  void setupEncoding(_xmlDoc* document) throw();

  /**
   * Si fuera necesario traduce el texto contenido por el parmetro recibido al conjunto
   * de caracteres indicado por el documento XML que estamos tratando.
   * \param source Texto expresado en el conjunto de caracteres particular.
   * \return Una cadena expresada en UTF-8.
   */
  const char* decode(const unsigned char* source) throw();

  /**
   * \warning Exclusivamente uso interno.
   */
  void attributes(Node* node, _xmlAttr* attribute) throw(RuntimeException);

  /**
   * \warning Exclusivamente uso interno.
   */
  void children(Node* node, _xmlNode* xmlNode) throw(RuntimeException);

private:
  _xmlValidCtxt* a_context;
  static char st_text [1024];
  static anna::RuntimeException* st_exception;
  bool a_encoding;
  DataBlock* a_buffEncode;

  Parser(const Parser&);

  void apply(_xmlDoc* document) throw(RuntimeException);

  static void errorHandler(void *ctx,  const char *msg, ...) throw();
  static void warningHandler(void *ctx,  const char *msg, ...) throw();
};

}
}

#endif
