// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_xml_XPath_hpp
#define anna_xml_XPath_hpp

#include <anna/core/RuntimeException.hpp>
#include <anna/core/mt/Mutex.hpp>
#include <anna/core/util/Recycler.hpp>

#include <anna/xml/Node.hpp>
#include <anna/xml/Parser.hpp>

struct _xmlNode;
struct _xmlNodeSet;
struct  _xmlXPathContext;

namespace anna {

namespace xml {

/**
   XPath is a language to navigate an XML document.

   \see https://www.w3schools.com/xml/xpath_intro.asp
   \see http://www.whoishostingthis.com/resources/xpath
*/
class XPath : Parser {
public:
  /**
   * Modo en que recupera la informacin de los nodos seleccionados por la expressin XPath.
   * \see XPath.
   */
  struct Mode {
    enum _v {
      Simple = 0x01, /**< Slo devuelve el primer nivel de nodos que cumplen con la expressin XPath. */
      Full = 0x02,  /**< Devuelve el rbol dependiente de los nodos que cumplen con la expressin XPath */
      Namespace = 0x04 /**< Activa el uso de namespaces en el doc XML recibido.*/
    };
  };

  /**
     Constructor.
     \param name Nombre lgico de esta instancia.
  */
  XPath(const char* name);

  /**
     Aplica la expresin XPath sobre el documento recibido como parmetro. Los nodos
     del documento XML seleccionados por XPath se recogeran con los mtodos
     #node_begin, #node_end y #node.

     \param document Documento XML sobre el que aplicar la expresin.
     \param expression Expresin XPath a aplicar sobre el documento XML.
     \param mode Una combinacin de XPath::Mode::_v.
     \warning La instancia del documento debe estar correctamente inicializada.
     \see Document::initialize
  */
  void apply(const Document& document, const char* expression, const int mode = Mode::Simple) throw(RuntimeException);

  /**
   * Aplica la expresin XPath sobre el documento recibido como parmetro y devuelve
   * \em true si hay algn nodo que cumpla la expresin o \em false en otro caso.
   * Es mucho ms rpido que el mtodo #apply, porque no dispondremos de los nodos que
   * cumplen la condicin.
   *
   * \param document Documento XML al que aplicar la sentencia.
   * \param expression Expresin XPath a aplicar.
   * \param mode Modo de interpretacin de los resultados. \em Si valor XPath::Mode::Namespace analizar el
   * documento XML para obtener los namespaces a usar.
   *
   * \return \em true si algn nodo del documento que cumpla la expressin o \em false en otro caso.
   */
  bool match(const Document& document, const char* expression, const int mode = Mode::Simple) throw(RuntimeException);

  /**
     Aplica la expresin XPath sobre el documento recibido como parmetro y devuelve el primer
     nodo que cumple con la expresin.
   *
   * \param document Documento XML al que aplicar la sentencia.
   * \param expression Expresin XPath a aplicar.
   * \param mode Modo de interpretacin de los resultados. \em Si valor XPath::Mode::Namespace analizar el
   * documento XML para obtener los namespaces a usar.
   * \param emode Modo de actuar en caso de que no haya ningn nodo que cumpla la expresin.
   *
   * \warning Si no hay ningn nodo que cumpla la expresin devolver una excepcin.
   */
  const Node* find(const Document& document, const char* expression, const int mode = Mode::Full, const Exception::Mode::_v emode = Exception::Mode::Throw)
  throw(RuntimeException);

  /**
   * Devuelve \em true si existen nodos que cumplieron la condicin de #apply o \em false en otro caso.
   * \return  \em true si existen nodos que cumplieron la condicin de #apply o \em false en otro caso.
   */
  bool isEmpty() const throw() { return a_root->a_children.size() == 0; }

  /**
   * Devuelve el nmero de nodos que han sido seleccionados al invocar al mtodo #apply.
   * \return El nmero de nodos que han sido seleccionados al invocar al mtodo #apply.
   */
  int size() const throw() { return a_root->a_children.size(); }

  /**
   * Devuelve el primer nodo del documento que cumple la expresin indicada al invocar al mtodo #apply.
   * \return El primer nodo del documento que cumple la expresin indicada al invocar al mtodo #apply.
   */
  Node::const_child_iterator node_begin() const throw() { return a_root->child_begin(); }

  /**
   * Devuelve el final de la lista de nodos que cumplen la expresin indicada al invocar al mtodo #apply.
   * \return El final de la lista de nodos que cumplen la expresin indicada al invocar al mtodo #apply.
   */
  Node::const_child_iterator node_end() const throw() { return a_root->child_end(); }

  /**
   * Devuelve la instancia del nodo apuntado por el iterador recibido como parmetro.
   * \return la instancia del nodo apuntado por el iterador recibido como parmetro.
   */
  static const Node* node(Node::const_child_iterator& ii) throw() { return xml::Node::node(ii); }

private:
  // Modo pasado al callback del apply
  int a_mode;

  // Resultado retornado por el callback del match
  bool a_result;

  typedef void(XPath::*Initialize)(_xmlXPathContext*, const Document&);
  typedef void (XPath::*Callback)(const _xmlNodeSet*);

  struct ConfigSkeleton;
  friend struct ConfigSkeleton;

  struct ConfigSkeleton {
    Initialize initialize;
    Callback callback;

    ConfigSkeleton(Callback _callback) : initialize(NULL), callback(_callback) {;}
  };

  XPath(const XPath&);

  void text(Node* node, _xmlNode* xmlNode) throw(RuntimeException);

  void skeleton(const Document&, const char* expression, ConfigSkeleton&) throw(RuntimeException);

  void initializeNamespaces(_xmlXPathContext*, const Document&) throw(RuntimeException);
  void forwardNamespaces(_xmlXPathContext*, _xmlNode*) throw(RuntimeException);
  void callbackApply(const _xmlNodeSet*);
  void callbackMatch(const _xmlNodeSet*);
};

}
}

#endif
