// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#include <anna/core/functions.hpp>
#include <anna/core/util/Variable.hpp>

using namespace std;
using namespace anna;

anna_assign_enum(Variable::Type) =  {
  "Unused0", "String", "Integer", "Unused1", "Boolean", "Integer64", "Block", "Float", "Double", "Custom", NULL
};

Variable::Variable(const char* name, const Type::_v type) :
  a_name(name),
  a_isNull(true),
  a_type(type),
  a_isOwner(true) {
  anna_memset(&a_value, 0, sizeof(a_value));

  switch(type) {
  case Type::String: a_value.a_string = new std::string; break;
  case Type::Integer: a_value.a_integer = &a_aux.integer; break;
  case Type::Integer64: a_value.a_longInteger = &a_aux.longInteger; break;
  case Type::Boolean: a_value.a_boolean = &a_aux.boolean; break;
  case Type::Block: a_value.a_dataBlock = new DataBlock(true); break;
  case Type::Float: a_value.a_float = &a_aux.theFloat; break;
  case Type::Double: a_value.a_double = &a_aux.theDouble; break;
  case Type::Custom: a_value.a_custom = NULL; break;
  default: break;
  }
}

Variable::~Variable() {
  if(a_isOwner == false)
    return;

  switch(a_type) {
  case Type::String:
    delete a_value.a_string;
    a_value.a_string = NULL;
    break;
  case Type::Block:
    delete a_value.a_dataBlock;
    a_value.a_dataBlock = NULL;
    break;
  default: break;
  }
}

void Variable::setValue(const char* value)
throw(RuntimeException) {
  verifyMatchType(Type::String, ANNA_FILE_LOCATION);
  a_isNull = false;
  *a_value.a_string = value;
}

void Variable::setValue(const int value)
throw(RuntimeException) {
  verifyMatchType(Type::Integer, ANNA_FILE_LOCATION);
  a_isNull = false;
  *a_value.a_integer = value;
}

void Variable::setValue(const S64 value)
throw(RuntimeException) {
  verifyMatchType(Type::Integer64, ANNA_FILE_LOCATION);
  a_isNull = false;
  *a_value.a_longInteger = value;
}

void Variable::setValue(const bool value)
throw(RuntimeException) {
  verifyMatchType(Type::Boolean, ANNA_FILE_LOCATION);
  a_isNull = false;
  *a_value.a_boolean = value;
}

void Variable::setValue(const DataBlock& value)
throw(RuntimeException) {
  verifyMatchType(Type::Block, ANNA_FILE_LOCATION);

  if(a_value.a_dataBlock->deepCopy() == false) {
    string msg("Variable: ");
    msg += a_name;
    msg += " | Datablock requires deep copy";
    throw RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  a_isNull = false;
  *a_value.a_dataBlock = value;
}

void Variable::setValue(const float value)
throw(RuntimeException) {
  verifyMatchSomeType(Type::Float, Type::Double, ANNA_FILE_LOCATION);

  if(a_type == Type::Float) {
    a_isNull = false;
    *a_value.a_float = value;
  } else if(a_type == Type::Double) {
    a_isNull = false;
    *a_value.a_double = value;
  }
}

void Variable::setValue(const double value)
throw(RuntimeException) {
  verifyMatchSomeType(Type::Float, Type::Double, ANNA_FILE_LOCATION);

  if(a_type == Type::Float) {
    a_isNull = false;
    *a_value.a_float = value;
  } else if(a_type == Type::Double) {
    a_isNull = false;
    *a_value.a_double = value;
  }
}

const char* Variable::getStringValue() const
throw(RuntimeException) {
  verifyIsNotNull(ANNA_FILE_LOCATION);
  verifyMatchType(Type::String, ANNA_FILE_LOCATION);
  return a_value.a_string->c_str();
}

int Variable::getIntegerValue() const
throw(RuntimeException) {
  verifyIsNotNull(ANNA_FILE_LOCATION);
  verifyMatchType(Type::Integer, ANNA_FILE_LOCATION);
  return *a_value.a_integer;
}

S64 Variable::getInteger64Value() const
throw(RuntimeException) {
  int result(0);
  verifyIsNotNull(ANNA_FILE_LOCATION);
  verifyMatchSomeType(Type::Integer, Type::Integer64, ANNA_FILE_LOCATION);

  switch(a_type) {
  case Type::Integer: result = *a_value.a_integer; break;
  case Type::Integer64: result = *a_value.a_longInteger; break;
  default: break;
  }

  return result;
}

bool Variable::getBooleanValue() const
throw(RuntimeException) {
  verifyIsNotNull(ANNA_FILE_LOCATION);
  verifyMatchType(Type::Boolean, ANNA_FILE_LOCATION);
  return *a_value.a_boolean;
}

const DataBlock& Variable::getDataBlockValue() const
throw(RuntimeException) {
  verifyIsNotNull(ANNA_FILE_LOCATION);
  verifyMatchType(Type::Block, ANNA_FILE_LOCATION);
  return *a_value.a_dataBlock;
}

float Variable::getFloatValue() const
throw(RuntimeException) {
  float result(0.0);
  verifyIsNotNull(ANNA_FILE_LOCATION);
  verifyMatchSomeType(Type::Float, Type::Double, ANNA_FILE_LOCATION);

  switch(a_type) {
  case Type::Float: result = *a_value.a_float; break;
  case Type::Double: result = (float) *a_value.a_double; break;
  default: break;
  }

  return result;
}

double Variable::getDoubleValue() const
throw(RuntimeException) {
  double result(0.0);
  verifyIsNotNull(ANNA_FILE_LOCATION);
  verifyMatchSomeType(Type::Float, Type::Double, ANNA_FILE_LOCATION);

  switch(a_type) {
  case Type::Float: result = (double) *a_value.a_float; break;
  case Type::Double: result = *a_value.a_double; break;
  default: break;
  }

  return result;
}

String Variable::asString() const
throw() {
  String result("anna::Variable { Name: ");
  result << a_name;
  result << " | Type: " << Type::asNotNullCString(a_type);
  result += " | Value: ";

  if(a_isNull == true)
    result += "<null>";
  else {
    switch(a_type) {
    case Type::String: result += a_value.a_string->c_str(); break;
    case Type::Integer: result += functions::asString(*a_value.a_integer); break;
    case Type::Integer64: result += functions::asString(*a_value.a_longInteger); break;
    case Type::Boolean: result += functions::asString(*a_value.a_boolean); break;
    case Type::Block: result += functions::asString(*a_value.a_dataBlock); break;
    case Type::Float: result += functions::asString("%f", *a_value.a_float); break;
    case Type::Double: result += functions::asString("%g", *a_value.a_double); break;
    case Type::Custom: result += functions::asHexString(anna_ptrnumber_cast(a_value.a_custom)); break;
    default: break;
    }
  }

  return result += " }";
}

void* Variable::buffer() const
throw() {
  void* result(NULL);

  switch(a_type) {
  case Type::String: result = const_cast <char*>(a_value.a_string->data()); break;
  case Type::Integer: result = a_value.a_integer; break;
  case Type::Integer64: result = a_value.a_longInteger; break;
  case Type::Boolean: result = a_value.a_boolean; break;
  case Type::Block: result = const_cast <char*>(a_value.a_dataBlock->getData()); break;
  case Type::Float: result = a_value.a_float; break;
  case Type::Double: result = a_value.a_double; break;
  case Type::Custom: result = a_value.a_custom; break;
  default: break;
  }

  return result;
}

void* Variable::getReference() const
throw() {
  void* result(NULL);

  switch(a_type) {
  case Type::String: result = a_value.a_string; break;
  case Type::Integer: result = a_value.a_integer; break;
  case Type::Integer64: result = a_value.a_longInteger; break;
  case Type::Boolean: result = a_value.a_boolean; break;
  case Type::Block: result = a_value.a_dataBlock; break;
  case Type::Float: result = a_value.a_float; break;
  case Type::Double: result = a_value.a_double; break;
  case Type::Custom: result = a_value.a_custom; break;
  default: break;
  }

  return result;
}

void Variable::verifyMatchType(const Type::_v type, const char* file, const int lineno) const
throw(RuntimeException) {
  if(a_type != type) {
    String msg("Variable: ");
    msg << a_name << " | " << Type::asNotNullCString(type) << " mismatch data type";
    throw RuntimeException(msg, file, lineno);
  }
}

void Variable::verifyMatchSomeType(const Type::_v firstType, const Type::_v secondType, const char* file, const int lineno) const
throw(RuntimeException) {
  if(a_type != firstType && a_type != secondType) {
    String msg("Variable: ");
    msg << a_name << " | Neihter " << Type::asNotNullCString(firstType);
    msg << " nor " << Type::asNotNullCString(secondType) << " matches with data type";
    throw RuntimeException(msg, file, lineno);
  }
}

void Variable::verifyIsNotNull(const char* file, const int lineno) const
throw(RuntimeException) {
  if(a_isNull == true) {
    String msg("Variable: ");
    msg << a_name << " | Variable does not have assigned value";
    throw RuntimeException(msg, file, lineno);
  }
}
